#use "ass12.ml"

(* Perform all possible beta-reductions on an expression until the normal form
 * has been reached *)
let rec eval exp =
    let get_int_value = function
          Num i -> i
        | _ -> raise (Failure "Not an integer")
    in
    let get_bool_value = function
          Bool b -> b
        | _ -> raise (Failure "Not a boolean")
    in
    let eval_monop op a = match op with
          Neg -> Num (- (get_int_value a))
        | Not -> Bool (not (get_bool_value a))
    in
    let eval_binop op x y = match op with
          And | Or ->
            let a = get_bool_value x in
            let b = get_bool_value y in
            (match op with
              And -> Bool (a && b)
            | Or -> Bool (a || b))
        | _ ->
            let a = get_int_value x in
            let b = get_int_value y in
            (match op with
              Add -> Num (a + b)
            | Sub -> Num (a - b)
            | Mul -> Num (a * b)
            | Div -> Num (a / b)
            | Eq -> Bool (a = b)
            | Ne -> Bool (a != b)
            | Lt -> Bool (a < b)
            | Le -> Bool (a <= b)
            | Gt -> Bool (a > b)
            | Ge -> Bool (a >= b))
    in
    match exp with
      MonopAp (op, e) -> eval_monop op (eval e)
    | BinopAp (op, e1, e2) -> eval_binop op (eval e1) (eval e2)
    | Fun (x, e) -> Fun (x, eval e)
    | FunAp (Fun (x, Var y), a) when y = x -> eval a (* identity function *)
    | FunAp (Fun (y, e), a) -> (* prevent recursion *)
        let result = (subs e y a) in
        if result = exp then result else eval result
    | FunAp (e1, e2) -> (* prevent recursion *)
        let result = FunAp (eval e1, eval e2) in
        if result = exp then result else eval result
    | Cond (cond, e1, e2) -> Cond (eval cond, eval e1, eval e2)
    | Let (x, e1, e2) -> eval (subs e2 x e1)
    | LetRec (x, e1, e2) -> LetRec (x, eval e1, eval e2)
    | _ -> exp
